from enum import Enum
from typing import Literal, Union


class AnalyzerProviderVendor(str, Enum):
    IBM_QUANTUM = "IBM Quantum"
    AZURE_QUANTUM = "Azure Quantum"
    AMAZON_BRAKET = "Amazon Braket"


class ProviderVendor(str, Enum):
    IBM_QUANTUM = "IBM Quantum"
    AZURE_QUANTUM = "Azure Quantum"
    AMAZON_BRAKET = "Amazon Braket"
    IONQ = "IonQ"
    NVIDIA = "Nvidia"


class ProviderTypeVendor:
    IBM_QUANTUM = Literal[ProviderVendor.IBM_QUANTUM]
    AZURE_QUANTUM = Literal[ProviderVendor.AZURE_QUANTUM]
    AWS_BRAKET = Literal[ProviderVendor.AMAZON_BRAKET]
    IONQ = Literal[ProviderVendor.IONQ]
    NVIDIA = Literal[ProviderVendor.NVIDIA]


class IonqBackendNames(str, Enum):
    SIMULATOR = "simulator"
    HARMONY = "qpu.harmony"
    ARIA = "qpu.aria-1"
    S11 = "qpu.s11"


class AzureQuantumBackendNames(str, Enum):
    IONQ_ARIA = "ionq.qpu.aria-1"
    IONQ_QPU = "ionq.qpu"
    IONQ_SIMULATOR = "ionq.simulator"
    MICROSOFT_ESTIMATOR = "microsoft.estimator"
    MICROSOFT_FULLSTATE_SIMULATOR = "microsoft.simulator.fullstate"
    RIGETTI_ASPEN3 = "rigetti.qpu.aspen-m-3"
    RIGETTI_SIMULATOR = "rigetti.sim.qvm"
    QCI_MACHINE1 = "qci.machine1"
    QCI_NOISY_SIMULATOR = "qci.simulator.noisy"
    QCI_SIMULATOR = "qci.simulator"
    QUANTINUUM_API_VALIDATOR1_1 = "quantinuum.sim.h1-1sc"
    QUANTINUUM_API_VALIDATOR1_1_OLD = "quantinuum.hqs-lt-s1-apival"
    QUANTINUUM_API_VALIDATOR1_2 = "quantinuum.sim.h1-2sc"
    QUANTINUUM_API_VALIDATOR1_2_OLD = "quantinuum.hqs-lt-s2-apival"
    QUANTINUUM_API_VALIDATOR2_1 = "quantinuum.sim.h2-1sc"
    QUANTINUUM_QPU1_1 = "quantinuum.qpu.h1-1"
    QUANTINUUM_QPU1_1_OLD = "quantinuum.hqs-lt-s1"
    QUANTINUUM_QPU1_2 = "quantinuum.qpu.h1-2"
    QUANTINUUM_QPU1_2_OLD = "quantinuum.hqs-lt-s2"
    QUANTINUUM_SIMULATOR1_1 = "quantinuum.sim.h1-1e"
    QUANTINUUM_SIMULATOR1_1_OLD = "quantinuum.hqs-lt-s1-sim"
    QUANTINUUM_SIMULATOR1_2 = "quantinuum.sim.h1-2e"
    QUANTINUUM_SIMULATOR1_2_OLD = "quantinuum.hqs-lt-s2-sim"
    QUANTINUUM_QPU2 = "quantinuum.qpu.h2-1"
    QUANTINUUM_SIMULATOR2 = "quantinuum.sim.h2-1e"


class AWSBackendNames(str, Enum):
    AMAZON_BRAKET_SV1 = "SV1"
    AMAZON_BRAKET_TN1 = "TN1"
    AMAZON_BRAKET_DM1 = "dm1"
    AMAZON_BRAKET_ASPEN_11 = "Aspen-11"
    AMAZON_BRAKET_M_1 = "Aspen-M-1"
    AMAZON_BRAKET_IONQ = "IonQ Device"
    AMAZON_BRAKET_LUCY = "Lucy"


class IBMQBackendNames(str, Enum):
    IBMQ_AER_SIMULATOR = "aer_simulator"
    IBMQ_AER_SIMULATOR_STATEVECTOR = "aer_simulator_statevector"
    IBMQ_AER_SIMULATOR_DENSITY_MATRIX = "aer_simulator_density_matrix"
    IBMQ_AER_SIMULATOR_MATRIX_PRODUCT_STATE = "aer_simulator_matrix_product_state"


class IBMQHardwareNames(str, Enum):
    ALMADEN = "Almaden"
    ARMONK = "Armonk"
    ATHENS = "Athens"
    BELEM = "Belem"
    BOEBLINGEN = "Boeblingen"
    BOGOTA = "Bogota"
    BROOKLYN = "Brooklyn"
    BURLINGTON = "Burlington"
    CAIRO = "Cairo"
    CAMBRIDGE = "Cambridge"
    CAMBRIDGEALTERNATIVEBASIS = "CambridgeAlternativeBasis"
    CASABLANCA = "Casablanca"
    ESSEX = "Essex"
    GUADALUPE = "Guadalupe"
    HANOI = "Hanoi"
    JAKARTA = "Jakarta"
    JOHANNESBURG = "Johannesburg"
    KOLKATA = "Kolkata"
    LAGOS = "Lagos"
    LIMA = "Lima"
    LONDON = "London"
    MANILA = "Manila"
    MANHATTAN = "Manhattan"
    MELBOURNE = "Melbourne"
    MONTREAL = "Montreal"
    MUMBAI = "Mumbai"
    NAIROBI = "Nairobi"
    OPENPULSE2Q = "OpenPulse2Q"
    OPENPULSE3Q = "OpenPulse3Q"
    OURENSE = "Ourense"
    PARIS = "Paris"
    POUGHKEEPSIE = "Poughkeepsie"
    QUITO = "Quito"
    ROCHESTER = "Rochester"
    ROME = "Rome"
    RUESCHLIKON = "Rueschlikon"
    SANTIAGO = "Santiago"
    SINGAPORE = "Singapore"
    SYDNEY = "Sydney"
    TENERIFE = "Tenerife"
    TOKYO = "Tokyo"
    TORONTO = "Toronto"
    VALENCIA = "Valencia"
    VIGO = "Vigo"
    WASHINGTON = "Washington"
    YORKTOWN = "Yorktown"


class NvidiaBackendNames(str, Enum):
    STATEVECTOR = "statevector"


EXACT_SIMULATORS = {
    IonqBackendNames.SIMULATOR,
    AzureQuantumBackendNames.IONQ_SIMULATOR,
    AzureQuantumBackendNames.MICROSOFT_FULLSTATE_SIMULATOR,
    AWSBackendNames.AMAZON_BRAKET_SV1,
    AWSBackendNames.AMAZON_BRAKET_TN1,
    AWSBackendNames.AMAZON_BRAKET_DM1,
    NvidiaBackendNames.STATEVECTOR,
    *IBMQBackendNames,
}

AllIBMQBackendNames = Union[IBMQBackendNames, IBMQHardwareNames]


class AllBackendsNameByVendor(str, Enum):
    IBM_QUANTUM = AllIBMQBackendNames
    AZURE_QUANTUM = AzureQuantumBackendNames
    AMAZON_BRAKET = AWSBackendNames
    IONQ = IonqBackendNames
    NVIDIA = NvidiaBackendNames
